import es6Promise from 'es6-promise';
import head from 'lodash/head';
import keys from 'lodash/keys';
import camelCase from 'lodash/camelCase';
import concat from 'lodash/concat';
import get from 'lodash/get';
import isArray from 'lodash/isArray';
import isEmpty from 'lodash/isEmpty';
import kebabCase from 'lodash/kebabCase';
import pullAll from 'lodash/pullAll';
import snakeCase from 'lodash/snakeCase';
import uniq from 'lodash/uniq';
import falcorPathParser from 'falcor-path-syntax';
import map from 'lodash/map';
import isNil from 'lodash/isNil';
import { stringify } from 'querystring';
import filter from 'lodash/filter';
import isError from 'lodash/isError';
import isFunction from 'lodash/isFunction';
import last from 'lodash/last';
import nth from 'lodash/nth';
import some from 'lodash/some';

var name = 'audioCategory';
var pluralName = 'audioCategories';

var defaultProperties = ['id', 'name'];

var audioCategory = Object.freeze({
	name: name,
	pluralName: pluralName,
	defaultProperties: defaultProperties
});

var name$1 = 'audioTrack';
var pluralName$1 = 'audioTracks';

var defaultProperties$1 = ['id', 'name', 'author', 'duration', 'src'];

var audioTrack = Object.freeze({
	name: name$1,
	pluralName: pluralName$1,
	defaultProperties: defaultProperties$1
});

var DEFAULT_FROM_INDEX = 0;
var DEFAULT_TO_INDEX = 19;

function buildPaging(options) {
  return {
    from: get(options, 'paging.from') || DEFAULT_FROM_INDEX,
    to: get(options, 'paging.to') || DEFAULT_TO_INDEX
  };
}

function buildProperties(defaults, options) {
  var properties = concat([], options.properties || defaults);

  var toAnd = options.andProperties;
  if (isArray(toAnd) && !isEmpty(toAnd)) {
    properties = concat(properties, toAnd);
  }

  var toNot = options.notProperties;
  if (isArray(toNot) && !isEmpty(toNot)) {
    properties = pullAll(properties, toNot);
  }

  properties = uniq(properties);

  return properties;
}

function entityNameToFilename(name) {
  return kebabCase(name) + '.js';
}

function entityToAPIEnvelopeKey(entity) {
  return snakeCase(entity.name);
}

function entityToFalcorListLeaf(entity) {
  return snakeCase(entity.name) + '_list';
}

function entityToFalcorRootLeaf(entity) {
  return snakeCase(entity.pluralName);
}

function falcorListLeafToEntityName(listLeaf) {
  return camelCase(listLeaf.replace(/_list$/, ''));
}

function falcorRootLeafToEntityName(rootLeaf) {
  return camelCase(rootLeaf);
}

function optionsHasQuery(options) {
  return keys(options.query).length > 0;
}

var name$2 = 'comment';
var pluralName$2 = 'comments';

function invalidateAfterCreate() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var parentEntityName = head(keys(options.parent));
  var parentEntity = entities[parentEntityName];
  var parentId = options.parent[parentEntityName].id;

  var rootLeaf = entityToFalcorRootLeaf(parentEntity);

  var paths = [[rootLeaf, parentId, 'comment_list'], [rootLeaf, parentId, 'comment_count']];

  return paths;
}

var comment = Object.freeze({
	name: name$2,
	pluralName: pluralName$2,
	invalidateAfterCreate: invalidateAfterCreate
});

var name$4 = 'post';
var pluralName$4 = 'posts';

var defaultProperties$2 = ['id', 'body'];

var post = Object.freeze({
	name: name$4,
	pluralName: pluralName$4,
	defaultProperties: defaultProperties$2
});

var name$3 = 'feedPost';
var pluralName$3 = 'feedPosts';
function getMany() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var _buildPaging = buildPaging(options),
      from = _buildPaging.from,
      to = _buildPaging.to;

  var properties = buildProperties(defaultProperties$2, options);
  var listLeaf = entityToFalcorListLeaf({ name: name$3 });

  var paths = map(properties, function (property) {
    var propertyPath = falcorPathParser(property);
    var path = ['self', listLeaf, { from: from, to: to }, propertyPath];

    return path;
  });

  return paths;
}

function extractMany() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  var jsonGraph = arguments[1];

  var listLeaf = entityToFalcorListLeaf({ name: name$3 });
  var postList = jsonGraph.json.self[listLeaf];

  var posts = map(postList, function (post) {
    return post;
  });

  return posts;
}

var feedPost = Object.freeze({
	name: name$3,
	pluralName: pluralName$3,
	defaultProperties: defaultProperties$2,
	getMany: getMany,
	extractMany: extractMany
});

var name$5 = 'moment';
var pluralName$5 = 'moments';

var defaultProperties$3 = ['id', 'description', 'title', 'user.name', 'cover_image_media.id', 'cover_image_media.src'];

var moment = Object.freeze({
	name: name$5,
	pluralName: pluralName$5,
	defaultProperties: defaultProperties$3
});

var name$7 = 'review';
var pluralName$7 = 'reviews';

var defaultProperties$4 = ['id', 'rating'];

var review = Object.freeze({
	name: name$7,
	pluralName: pluralName$7,
	defaultProperties: defaultProperties$4
});

var name$8 = 'tree';
var pluralName$8 = 'trees';

var treeEntity = Object.freeze({
	name: name$8,
	pluralName: pluralName$8
});

/*
 * This entity is temporary while we figure out how to handle "my" entities
 * in a standard entity way. We expect to delete it once we have settled on a
 * standard "my" approach.
 */

var name$6 = 'myTreeReview';
var pluralName$6 = 'myTreeReviews';
function getOne() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  var tree = options.tree;


  if (isNil(tree) || isNil(tree.id)) {
    return new Error("options.tree.id is required when getting the current user's tree review");
  }

  var properties = buildProperties(defaultProperties$4, options);

  var paths = properties.map(function (property) {
    var rootLeaf = entityToFalcorRootLeaf(treeEntity);
    var path = falcorPathParser(property);

    return concat([rootLeaf, tree.id, 'my_review'], path);
  });

  return paths;
}

function extractOne() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  var jsonGraph = arguments[1];
  var tree = options.tree;


  if (isNil(tree) || isNil(tree.id)) {
    return new Error("options.tree.id is required when getting the current user's tree review");
  }

  var rootLeaf = entityToFalcorRootLeaf(treeEntity);

  return jsonGraph.json[rootLeaf][tree.id].my_review;
}

var myTreeReview = Object.freeze({
	name: name$6,
	pluralName: pluralName$6,
	defaultProperties: defaultProperties$4,
	getOne: getOne,
	extractOne: extractOne
});

var name$9 = 'note';
var pluralName$9 = 'notes';

var defaultProperties$5 = ['id', 'body'];

var note = Object.freeze({
	name: name$9,
	pluralName: pluralName$9,
	defaultProperties: defaultProperties$5
});

var name$10 = 'space';
var pluralName$10 = 'spaces';

var space = Object.freeze({
	name: name$10,
	pluralName: pluralName$10
});

var name$11 = 'timeline';
var pluralName$11 = 'timelines';

var timeline = Object.freeze({
	name: name$11,
	pluralName: pluralName$11
});

var name$12 = 'topic';
var pluralName$12 = 'topics';

var defaultProperties$6 = ['id', 'name'];

var topic = Object.freeze({
	name: name$12,
	pluralName: pluralName$12,
	defaultProperties: defaultProperties$6
});

var name$13 = 'tourTipView';
var pluralName$13 = 'tourTipViews';

var defaultProperties$7 = ['id', 'state'];

function getMany$1() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var _buildPaging = buildPaging(options),
      from = _buildPaging.from,
      to = _buildPaging.to;

  var properties = buildProperties(defaultProperties$7, options);
  var listLeaf = entityToFalcorListLeaf({ name: name$13 });

  var paths = map(properties, function (property) {
    var propertyPath = falcorPathParser(property);
    var path = ['self', listLeaf, { from: from, to: to }, propertyPath];

    return path;
  });

  return paths;
}

function extractMany$1() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  var jsonGraph = arguments[1];

  var listLeaf = entityToFalcorListLeaf({ name: name$13 });
  var viewList = jsonGraph.json.self[listLeaf];

  var views = map(viewList, function (view) {
    return view;
  });

  return views;
}

var tourTipView = Object.freeze({
	name: name$13,
	pluralName: pluralName$13,
	defaultProperties: defaultProperties$7,
	getMany: getMany$1,
	extractMany: extractMany$1
});



var entities = Object.freeze({
	audioCategory: audioCategory,
	audioTrack: audioTrack,
	comment: comment,
	feedPost: feedPost,
	moment: moment,
	myTreeReview: myTreeReview,
	note: note,
	post: post,
	review: review,
	space: space,
	timeline: timeline,
	topic: topic,
	tourTipView: tourTipView,
	tree: treeEntity
});

var entitySrc = 'entities';

function entityIsUndefinedError(entityName) {
  var filename = entityNameToFilename(entityName);
  return new Error('The ' + entityName + ' entity is undefined. Ensure ' + entitySrc + '/' + filename + ' exists and is exported by ' + entitySrc + '/_catalog.js');
}

function entityMustExportError(entityName, property) {
  var filename = entityNameToFilename(entityName);
  return new Error(entitySrc + '/' + filename + ' must export ' + property);
}

function optionsMustHaveError(property) {
  return new Error('options.' + property + ' is required');
}

function entityExports(entity, property) {
  return !isNil(entity[property]);
}



function optionsHas(options, property) {
  return !isNil(get(options, property));
}

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function getOne$1(entityName, entity) {
  var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  if (!entityExports(entity, 'pluralName')) {
    return entityMustExportError(entityName, 'pluralName');
  }

  if (!entityExports(entity, 'defaultProperties')) {
    return entityMustExportError(entityName, 'defaultProperties');
  }

  if (!optionsHas(options, 'id')) {
    return optionsMustHaveError('id');
  }

  var properties = buildProperties(entity.defaultProperties, options);
  var rootLeaf = entityToFalcorRootLeaf(entity);

  var paths = properties.map(function (property) {
    var propertyPath = falcorPathParser(property);
    var path = concat([rootLeaf, options.id], propertyPath);

    return path;
  });

  return paths;
}

function extractOne$1(entity) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  var jsonGraph = arguments[2];

  var rootLeaf = entityToFalcorRootLeaf(entity);
  var extractedEntity = jsonGraph.json[rootLeaf][options.id];

  return extractedEntity;
}

function getMany$2(entityName, entity) {
  var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  if (!entityExports(entity, 'name')) {
    return entityMustExportError(entityName, 'name');
  }

  if (!entityExports(entity, 'pluralName')) {
    return entityMustExportError(entityName, 'pluralName');
  }

  if (!entityExports(entity, 'defaultProperties')) {
    return entityMustExportError(entityName, 'defaultProperties');
  }

  var _buildPaging = buildPaging(options),
      from = _buildPaging.from,
      to = _buildPaging.to;

  var properties = buildProperties(entity.defaultProperties, options);

  var parentEntity = void 0,
      parentOptions = void 0;
  if (!isNil(options.parent)) {
    var parentEntityName = head(keys(options.parent));
    parentEntity = entities[parentEntityName];
    parentOptions = options.parent[parentEntityName];
  }

  var basePath = void 0;
  if (parentEntity) {
    var rootLeaf = entityToFalcorRootLeaf(parentEntity);
    var listLeaf = entityToFalcorListLeaf(entity);

    basePath = [rootLeaf, parentOptions.id, listLeaf];
  } else {
    var _listLeaf = entityToFalcorListLeaf(entity);

    basePath = [_listLeaf];
  }

  if (optionsHasQuery(options)) {
    var qs = stringify(options.query);
    basePath.push(qs);
  }

  var paths = properties.map(function (property) {
    var propertyPath = falcorPathParser(property);
    var path = concat(basePath, { from: from, to: to }, propertyPath);

    return path;
  });

  return paths;
}

function extractMany$2(entity) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  var jsonGraph = arguments[2];

  var parentEntity = void 0,
      parentOptions = void 0;
  if (!isNil(options.parent)) {
    var parentEntityName = head(keys(options.parent));
    parentEntity = entities[parentEntityName];
    parentOptions = options.parent[parentEntityName];
  }

  var entityList = void 0;
  if (parentEntity) {
    var parentRootLeaf = entityToFalcorRootLeaf(parentEntity);
    var entityListLeaf = entityToFalcorListLeaf(entity);

    entityList = jsonGraph.json[parentRootLeaf][parentOptions.id][entityListLeaf];
  } else {
    var listLeaf = entityToFalcorListLeaf(entity);

    entityList = jsonGraph.json[listLeaf];
  }

  if (optionsHasQuery(options)) {
    var qs = stringify(options.query);
    entityList = entityList[qs];
  }

  var extractedEntities = map(entityList, function (entity) {
    return entity;
  });

  return extractedEntities;
}

function create(entityName, entity) {
  var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  if (!entityExports(entity, 'name')) {
    return entityMustExportError(entityName, 'name');
  }

  if (!entityExports(entity, 'pluralName')) {
    return entityMustExportError(entityName, 'pluralName');
  }

  if (!optionsHas(options, entityName)) {
    return optionsMustHaveError(entityName);
  }

  var parentEntity = void 0,
      parentOptions = void 0;
  if (!isNil(options.parent)) {
    var parentEntityName = head(keys(options.parent));
    parentEntity = entities[parentEntityName];
    parentOptions = options.parent[parentEntityName];

    if (isNil(parentEntity)) {
      return entityIsUndefinedError(parentEntityName);
    }

    if (!entityExports(parentEntity, 'pluralName')) {
      return entityMustExportError(parentEntityName, 'pluralName');
    }

    if (!optionsHas(parentOptions, 'id')) {
      return optionsMustHaveError('parent.' + parentEntityName + '.id');
    }
  }

  var path = void 0;
  if (parentEntity) {
    var rootLeaf = entityToFalcorRootLeaf(parentEntity);
    var listLeaf = entityToFalcorListLeaf(entity);

    path = [rootLeaf, parentOptions.id, listLeaf, 'add'];
  } else {
    var _rootLeaf = entityToFalcorRootLeaf(entity);

    path = [_rootLeaf, 'create'];
  }

  var envelopeKey = entityToAPIEnvelopeKey(entity);
  var args = [_defineProperty({}, envelopeKey, options[entity.name])];

  return { path: path, args: args };
}

function extractCreated(entity) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  var jsonGraph = arguments[2];

  var rootLeaf = entityToFalcorRootLeaf(entity);
  var entities = jsonGraph.json[rootLeaf];
  var first = head(keys(entities));
  var extractedEntity = entities[first];

  return extractedEntity;
}

function invalidateAfterCreate$1(entity) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  if (!isNil(options.parent)) {
    var parentEntityName = head(keys(options.parent));
    var parentEntity = entities[parentEntityName];
    var parentId = options.parent[parentEntityName].id;

    var rootLeaf = entityToFalcorRootLeaf(parentEntity);
    var listLeaf = entityToFalcorListLeaf(entity);

    var paths = [[rootLeaf, parentId, listLeaf], [rootLeaf, parentId, listLeaf, 'length']];

    return paths;
  }
}

function update(entityName, entity) {
  var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  if (!entityExports(entity, 'pluralName')) {
    return entityMustExportError(entityName, 'pluralName');
  }

  if (!optionsHas(options, 'id')) {
    return optionsMustHaveError('id');
  }

  if (keys(options).length === 1) {
    return new Error('options must include at least one property to update');
  }

  var properties = filter(keys(options), function (key) {
    return key !== 'id';
  });
  var rootLeaf = entityToFalcorRootLeaf(entity);

  var paths = map(properties, function (property) {
    var path = {
      path: [rootLeaf, options.id, property],
      value: options[property]
    };

    return path;
  });

  return paths;
}

function extractUpdated(entity) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  var jsonGraph = arguments[2];

  var rootLeaf = entityToFalcorRootLeaf(entity);
  var updatedProperties = jsonGraph.json[rootLeaf][options.id];

  return updatedProperties;
}

function destroy(entityName, entity) {
  var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  if (!entityExports(entity, 'pluralName')) {
    return entityMustExportError(entityName, 'pluralName');
  }

  if (!optionsHas(options, 'id')) {
    return optionsMustHaveError('id');
  }

  var rootLeaf = entityToFalcorRootLeaf(entity);
  var destroyPath = [rootLeaf, options.id, 'delete'];

  return destroyPath;
}

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

es6Promise.polyfill();

var SuperGloo = function () {
  function SuperGloo() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

    _classCallCheck(this, SuperGloo);

    var requiredProperties = ['falcorModel'];

    if (some(requiredProperties, function (_) {
      return isNil(get(config, _));
    })) {
      throw new Error('One of the following required config parameters is undefined: ' + requiredProperties.join(', '));
    }

    this.falcor = config.falcorModel;
  }

  /*
   * Standard interfaces
   */

  _createClass(SuperGloo, [{
    key: 'setModel',
    value: function setModel(model) {
      if (isNil(model)) {
        throw new Error('When setting the model, the model cannot be null or undefined');
      }

      this.falcor = model;
    }
  }, {
    key: 'getMany',
    value: function getMany$$1(entityName) {
      var _falcor;

      var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      var entity = entities[entityName];

      if (isNil(entity)) {
        return Promise.reject(entityIsUndefinedError(entityName));
      }

      var paths = isFunction(entity.getMany) ? entity.getMany(options) : getMany$2(entityName, entity, options);

      if (isError(paths)) {
        return Promise.reject(paths);
      }

      return (_falcor = this.falcor).get.apply(_falcor, _toConsumableArray(paths)).then(function (jsonGraph) {
        var extractedEntities = void 0;
        if (isNil(get(jsonGraph, 'json'))) {
          extractedEntities = null;
        } else {
          extractedEntities = isFunction(entity.extractMany) ? entity.extractMany(options, jsonGraph) : extractMany$2(entity, options, jsonGraph);
        }

        return extractedEntities;
      }).catch(function (response) {
        var message = buildGetManyErrorMessage(response);
        return Promise.reject(new Error(message));
      });
    }
  }, {
    key: 'getOne',
    value: function getOne$$1(entityName) {
      var _falcor2;

      var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      var entity = entities[entityName];

      if (isNil(entity)) {
        return Promise.reject(entityIsUndefinedError(entityName));
      }

      var paths = isFunction(entity.getOne) ? entity.getOne(options) : getOne$1(entityName, entity, options);

      if (isError(paths)) {
        return Promise.reject(paths);
      }

      return (_falcor2 = this.falcor).get.apply(_falcor2, _toConsumableArray(paths)).then(function (jsonGraph) {
        var extractedEntity = void 0;
        if (isNil(get(jsonGraph, 'json'))) {
          extractedEntity = null;
        } else {
          extractedEntity = isFunction(entity.extractOne) ? entity.extractOne(options, jsonGraph) : extractOne$1(entity, options, jsonGraph);
        }

        return extractedEntity;
      }).catch(function (response) {
        var message = buildGetOneErrorMessage(response);
        return Promise.reject(new Error(message));
      });
    }
  }, {
    key: 'create',
    value: function create$$1(entityName) {
      var _this = this;

      var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      var entity = entities[entityName];

      if (isNil(entity)) {
        return Promise.reject(entityIsUndefinedError(entityName));
      }

      var callOptions = isFunction(entity.create) ? entity.create(options) : create(entityName, entity, options);

      if (isError(callOptions)) {
        return Promise.reject(callOptions);
      }

      var path = callOptions.path,
          args = callOptions.args;


      return this.falcor.call(path, args).then(function (jsonGraph) {
        var extractedEntity = void 0;
        if (isNil(get(jsonGraph, 'json'))) {
          extractedEntity = null;
        } else {
          extractedEntity = isFunction(entity.extractCreated) ? entity.extractCreated(options, jsonGraph) : extractCreated(entity, options, jsonGraph);
        }

        var pathsToInvalidate = isFunction(entity.invalidateAfterCreate) ? entity.invalidateAfterCreate(options) : invalidateAfterCreate$1(entity, options);

        if (pathsToInvalidate) {
          var _falcor3;

          (_falcor3 = _this.falcor).invalidate.apply(_falcor3, _toConsumableArray(pathsToInvalidate));
        }

        return extractedEntity;
      }).catch(function (response) {
        var message = buildCreateErrorMessage(response);
        return Promise.reject(new Error(message));
      });
    }
  }, {
    key: 'update',
    value: function update$$1(entityName) {
      var _falcor4;

      var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      var entity = entities[entityName];

      if (isNil(entity)) {
        return Promise.reject(entityIsUndefinedError(entityName));
      }

      var paths = isFunction(entity.update) ? entity.update(options) : update(entityName, entity, options);

      if (isError(paths)) {
        return Promise.reject(paths);
      }

      return (_falcor4 = this.falcor).set.apply(_falcor4, _toConsumableArray(paths)).then(function (jsonGraph) {
        var updatedProperties = void 0;
        if (isNil(get(jsonGraph, 'json'))) {
          updatedProperties = null;
        } else {
          updatedProperties = isFunction(entity.extractUpdated) ? entity.extractUpdated(options, jsonGraph) : extractUpdated(entity, options, jsonGraph);
        }

        return updatedProperties;
      }).catch(function (response) {
        var message = buildUpdateErrorMessage(response);
        return Promise.reject(new Error(message));
      });
    }
  }, {
    key: 'destroy',
    value: function destroy$$1(entityName) {
      var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      var entity = entities[entityName];

      if (isNil(entity)) {
        return Promise.reject(entityIsUndefinedError(entityName));
      }

      var destroyPath = isFunction(entity.destroy) ? entity.destroy(options) : destroy(entityName, entity, options);

      if (isError(destroyPath)) {
        return Promise.reject(destroyPath);
      }

      return this.falcor.call(destroyPath).then(function () {
        return 'Entity destroyed';
      }).catch(function (response) {
        var message = buildDestroyErrorMessage(response);
        return Promise.reject(new Error(message));
      });
    }
  }]);

  return SuperGloo;
}();

var codeToMessageMap = {
  400: 'request is malformed',
  401: 'user is not authorized',
  403: 'user is forbidden',
  404: 'entity does not exist',
  422: 'entity is not valid',
  default: 'unspecified error encountered'
};

function buildGetOneErrorMessage(response) {
  if (response instanceof Error) {
    return response.message;
  }

  var firstPath = response[0];
  var entity = camelCase(firstPath.path[0]);
  var entityId = firstPath.path[1];
  var originalMessage = firstPath.value;

  var code = originalMessage.match(/\d{3}/)[0];
  var improvedMessage = codeToMessageMap[code] || codeToMessageMap.default;

  return 'Could not get ' + entity + ' ' + entityId + ': ' + improvedMessage;
}

function buildGetManyErrorMessage(response) {
  if (response instanceof Error) {
    return response.message;
  }

  var firstPath = response[0];
  var entity = camelCase(firstPath.path[0]);
  var originalMessage = firstPath.value;

  var code = originalMessage.match(/\d{3}/)[0];
  var improvedMessage = codeToMessageMap[code] || codeToMessageMap.default;

  return 'Could not get ' + entity + ': ' + improvedMessage;
}

function buildCreateErrorMessage(response) {
  if (response instanceof Error) {
    return response.message;
  }

  var firstError = head(response);

  var entityName = void 0;
  if (last(firstError.path) === 'add') {
    var listLeaf = nth(firstError.path, -2);
    entityName = falcorListLeafToEntityName(listLeaf);
  } else {
    var rootLeaf = head(firstError.path);
    entityName = falcorRootLeafToEntityName(rootLeaf);
  }

  var originalMessage = firstError.value;
  var code = originalMessage.match(/\d{3}/)[0];
  var improvedMessage = codeToMessageMap[code] || codeToMessageMap.default;

  return 'Could not create a new ' + entityName + ': ' + improvedMessage + ' (' + originalMessage + ')';
}

function buildUpdateErrorMessage(response) {
  if (response instanceof Error) {
    return response.message;
  }

  var firstPath = response[0];
  var entity = camelCase(firstPath.path[0]);
  var entityId = firstPath.path[1];
  var originalMessage = firstPath.value;

  var code = originalMessage.match(/\d{3}/)[0];
  var improvedMessage = codeToMessageMap[code] || codeToMessageMap.default;

  return 'Could not update ' + entity + ' ' + entityId + ': ' + improvedMessage + ' (' + originalMessage + ')';
}

function buildDestroyErrorMessage(response) {
  if (response instanceof Error) {
    return response.message;
  }

  var firstPath = response[0];
  var entity = camelCase(firstPath.path[0]);
  var entityId = firstPath.path[1];
  var originalMessage = firstPath.value;

  var code = originalMessage.match(/\d{3}/)[0];
  var improvedMessage = codeToMessageMap[code] || codeToMessageMap.default;

  return 'Could not destroy ' + entity + ' ' + entityId + ': ' + improvedMessage + ' (' + originalMessage + ')';
}

export default SuperGloo;
